/// ReWindowLibrary
/// A C++20 Library for creating and managing windows in a platform-independent manner
/// Developed and Maintained by the boys @ Redacted Software.
/// (c) 2024 Redacted Software
/// This work is dedicated to the public domain.

/// @file keyboard.h
/// @desc A class that models the functionality of a mouse / pointer device.
/// @edit 2024-07-29

#pragma once


namespace ReWindow
{
    class MouseState {
    public:
        struct
        {
            bool LMB = false;
            bool RMB = false;
            bool MMB = false;
            bool SideButton1 = false;
            bool SideButton2 = false;
            bool MWheelUp = false;
            bool MWheelDown = false;
        } Buttons;


        Vector2 Position;
        int Wheel = 0;

        [[nodiscard]] bool IsDown(const MouseButton& btn) const
        {
            if (btn == MouseButtons::Left) return Buttons.LMB;
            if (btn == MouseButtons::Right) return Buttons.RMB;
            if (btn == MouseButtons::Middle) return Buttons.MMB;
            if (btn == MouseButtons::Mouse4) return Buttons.SideButton1;
            if (btn == MouseButtons::Mouse5) return Buttons.SideButton2;
            //if (btn == MouseButtons::MWheelUp) return Buttons.MWheelUp;
            //if (btn == MouseButtons::MWheelDown) return Buttons.MWheelDown;

            return false; // Unknown button?
        }

        void Set(const MouseButton& btn, bool state)
        {
            if (btn == MouseButtons::Left) Buttons.LMB = state;
            if (btn == MouseButtons::Right) Buttons.RMB = state;
            if (btn == MouseButtons::Middle) Buttons.MMB = state;
            if (btn == MouseButtons::Mouse4) Buttons.SideButton1 = state;
            if (btn == MouseButtons::Mouse5) Buttons.SideButton2 = state;
            //if (btn == MouseButtons::MWheelUp) Buttons.MWheelUp = state;
            //if (btn == MouseButtons::MWheelDown) Buttons.MWheelDown = state;
        }

        bool& operator[](const MouseButton& btn)
        {
            if (btn == MouseButtons::Left) return Buttons.LMB;
            if (btn == MouseButtons::Right) return Buttons.RMB;
            if (btn == MouseButtons::Middle) return Buttons.MMB;
            if (btn == MouseButtons::Mouse4) return Buttons.SideButton1;
            if (btn == MouseButtons::Mouse5) return Buttons.SideButton2;
            //if (btn == MouseButtons::MWheelUp) return Buttons.MWheelUp;
            //if (btn == MouseButtons::MWheelDown) return Buttons.MWheelDown;

            throw std::runtime_error("Attempted to handle unmapped mouse button");
        }
    };

    class InputDevice {}; // TODO: Remember to break InputDevice into it's own file and not define it twice!!!

    class Pointer : public InputDevice {};


    class Mouse : public Pointer
    {

    };
}