#pragma once

namespace ReWindow { class RWindowImpl; }

class ReWindow::RWindowImpl {
protected: // Should maybe make private
    int width = 1280;
    int height = 720;

    bool open = false; // Is the underlying OS-Window-Handle actually open.
    bool resizable = true;
    bool fullscreen_mode = false;
    bool focused = true;
    bool vsync = false;
    bool cursor_visible = true;
    bool closing = false;

    float delta_time = 0.f;
    float refresh_rate = 0.f;
    unsigned int refresh_count = 0;

    std::string title = "Redacted Window";

    Vector2 lastKnownWindowSize {0, 0};

    // TODO: Implement ringbuffer / circular vector class of some sort.
    float refresh_rate_prev_1 = 0.f;
    float refresh_rate_prev_2 = 0.f;
    float refresh_rate_prev_3 = 0.f;
    float refresh_rate_prev_4 = 0.f;
    float refresh_rate_prev_5 = 0.f;

    float avg_refresh_rate = 0.0f;

    // Figure out what to do with this shit later
    bool flags[5];
    KeyboardState currentKeyboard;  // current frame keyboard state.
    KeyboardState previousKeyboard; // previous frame keyboard state.
    MouseState currentMouse;		// purrent frame mouse state.
    MouseState previousMouse;		// previous frame mouse state
    RenderingAPI renderer = RenderingAPI::OPENGL;
    //
public:
    explicit RWindowImpl(const RWindowParams& params)
        : RWindowImpl(params.title, params.width, params.height, RenderingAPI::OPENGL, params.fullscreen, p)
    explicit RWindowImpl(const std::string& wTitle,
                         int wWidth = 640, int wHeight = 480,
                         RenderingAPI wRenderer = RenderingAPI::OPENGL,
                         bool wFullscreen = false,
                         bool wResizable = true,
                         bool wVsync = false);
    ~RWindowImpl();
public:
    virtual void Open();
    virtual void Close();
    virtual void PollEvents();
    virtual void Refresh();
public:
    bool IsFocused() const;
    bool IsFullscreen() const;
    bool IsResizable() const;
    bool IsVsyncEnabled() const;
    bool IsAlive() const;
public:
    void SetFullscreen(bool fs);
    void SetResizable(bool fs);
    void SetVsyncEnabled(bool vsync);
    void SetSize(int width, int height);
    void SetSize(const Vector2& size);
    // Added here for now
    void SetPos(int x, int y);
    void SetPos(const Vector2& pos);
    //
    virtual void SetTitle(const std::string& title);
public:
    std::string GetTitle() const;
    Vector2 GetPos() const;
    Vector2 GetSize() const;
    int GetWidth() const;
    int GetHeight() const;
    float GetDeltaTime() const;
    float GetRefreshRate() const;
    float GetRefreshCounter() const;
public:
    // Figure out what to do with these later
    void Raise() const;
    void Lower() const;
    void DestroyOSWindowHandle();
    void SetCursorVisible(bool cursor_enable);
    Vector2 GetAccurateMouseCoordinates() const;
    void GLSwapBuffers();
    void Fullscreen();
    void RestoreFromFullscreen();
    // I know this doesn't modify the class, but it indirectly modifies the window
    // Making it const just seems deceptive.
    void SetCursorStyle(CursorStyle style) const;
    Vector2 GetPositionOfRenderableArea() const;
    std::string getGraphicsDriverVendor();
    void SetFlag(RWindowFlags flag, bool state);
    // Make protected
    void processKeyRelease (Key key);
    void processKeyPress (Key key);
    void processOnClose();
    void processOnOpen();
    void processMousePress(const MouseButton& btn);
    void processMouseRelease(const MouseButton& btn);
    void processFocusIn();
    void processFocusOut();
    void processMouseMove(Vector2 last_pos, Vector2 new_pos);
    void processMouseWheel(int scrolls);
    //
    virtual void OnOpen() {}
    virtual void OnClosing() {}
    virtual void OnFocusLost(const RWindowEvent& e) {}
    virtual void OnFocusGain(const RWindowEvent& e) {}
    virtual void OnRefresh(float elapsed) {}
    virtual void OnResizeSuccess() {}
    virtual bool OnResizeRequest(const WindowResizeRequestEvent& e) { return true;}
    virtual void OnKeyDown(const KeyDownEvent&) {}
    virtual void OnKeyUp(const KeyUpEvent&) {}
    virtual void OnMouseMove(const MouseMoveEvent&) {}
    virtual void OnMouseButtonDown(const MouseButtonDownEvent&) {}
    virtual void OnMouseButtonUp(const MouseButtonUpEvent&) {}
    virtual void OnMouseWheel(const MouseWheelEvent&) {}
    Event<> OnOpenEvent;
    Event<> OnClosingEvent;
    Event<RWindowEvent> OnFocusLostEvent;
    Event<RWindowEvent> OnFocusGainEvent;
    Event<float> OnRefreshEvent;
    Event<WindowResizeRequestEvent> OnResizeRequestEvent;
    Event<KeyDownEvent> OnKeyDownEvent;
    Event<KeyUpEvent> OnKeyUpEvent;
    Event<MouseMoveEvent> OnMouseMoveEvent;
    Event<MouseButtonDownEvent> OnMouseButtonDownEvent;
    Event<MouseButtonUpEvent> OnMouseButtonUpEvent;
    Event<MouseWheelEvent> OnMouseWheelEvent;

    Vector2 GetMouseCoordinates() const;
    bool GetFlag(RWindowFlags flag) const;
    //bool IsAlive() const;
    void SetSizeWithoutEvent(const Vector2& size);
    std::vector<RWindowEvent> eventLog;
    std::queue<RWindowEvent> eventQueue;
    void LogEvent(const RWindowEvent& e) { eventLog.push_back(e);};
    RWindowEvent GetLastEvent() const {
        return eventLog.back();
    };
    void SetLastKnownWindowSize(const Vector2& size);
    void SetRenderer(RenderingAPI api);
    bool IsKeyDown(Key key) const;
    bool IsMouseButtonDown(const MouseButton &button) const;
    void ManagedRefresh();
    float ComputeElapsedFrameTimeSeconds(std::chrono::steady_clock::time_point start, std::chrono::steady_clock::time_point end);
    std::chrono::steady_clock::time_point GetTimestamp();
    void UpdateFrameTiming(float frame_time);
    /*
    bool IsResizable() const;
    bool IsFullscreen() const;
    bool IsFocused() const;
    bool IsVsyncEnabled() const;
     */
    void ForceClose();
    void ForceCloseAndTerminateProgram();
    int GetMouseWheelPersistent() const { return currentMouse.Wheel;}
    [[nodiscard]] bool IsOpen() const { return open;}
    [[nodiscard]] bool IsClosing() const { return closing;}
};
